@@ -33,38 +33,45 @@
 #include "serial.h"
 #include "stm32.h"
 #include "parsers/parser.h"
+#include "port.h"
 
 #include "parsers/binary.h"
 #include "parsers/hex.h"
 
-#define VERSION "0.3"
+#define VERSION "0.4"
 
 /* device globals */
-serial_t	*serial		= NULL;
 stm32_t		*stm		= NULL;
 
 void		*p_st		= NULL;
 parser_t	*parser		= NULL;
 
 /* settings */
-char		*device		= NULL;
-serial_baud_t	baudRate	= SERIAL_BAUD_57600;
-char		*serial_mode	= "8e1";
+struct port_options port_opts = {
+	.device			= NULL,
+	.baudRate		= SERIAL_BAUD_57600,
+	.serial_mode		= "8e1",
+	.bus_addr		= 0,
+	.rx_frame_max		= STM32_MAX_RX_FRAME,
+	.tx_frame_max		= STM32_MAX_TX_FRAME,
+};
 int		rd	 	= 0;
 int		wr		= 0;
 int		wu		= 0;
 int		rp		= 0;
 int		ur		= 0;
 int		eraseOnly	= 0;
+int		crc		= 0;
 int		npages		= 0;
 int             spage           = 0;
+int             no_erase        = 0;
 char		verify		= 0;
 int		retry		= 10;
 char		exec_flag	= 0;
 uint32_t	execute		= 0;
 char		init_flag	= 1;
 char		force_binary	= 0;
-char		reset_flag	= 1;
+char		reset_flag	= 0;
 char		*filename;
 char		*gpio_seq	= NULL;
 uint32_t	start_addr	= 0;
@@ -74,8 +81,42 @@ uint32_t	readwrite_len	= 0;
 int  parse_options(int argc, char *argv[]);
 void show_help(char *name);
 
+static int is_addr_in_ram(uint32_t addr)
+{
+	return addr >= stm->dev->ram_start && addr < stm->dev->ram_end;
+}
+
+static int is_addr_in_flash(uint32_t addr)
+{
+	return addr >= stm->dev->fl_start && addr < stm->dev->fl_end;
+}
+
+static int flash_addr_to_page_floor(uint32_t addr)
+{
+	if (!is_addr_in_flash(addr))
+		return 0;
+
+	return (addr - stm->dev->fl_start) / stm->dev->fl_ps;
+}
+
+static int flash_addr_to_page_ceil(uint32_t addr)
+{
+	if (!(addr >= stm->dev->fl_start && addr <= stm->dev->fl_end))
+		return 0;
+
+	return (addr + stm->dev->fl_ps - 1 - stm->dev->fl_start)
+	       / stm->dev->fl_ps;
+}
+
+static uint32_t flash_page_to_addr(int page)
+{
+	return stm->dev->fl_start + page * stm->dev->fl_ps;
+}
+
 int main(int argc, char* argv[]) {
+	struct port_interface *port = NULL;
 	int ret = 1;
+	stm32_err_t s_err;
 	parser_err_t perr;
 	FILE *diag = stdout;
 
@@ -134,31 +175,23 @@ int main(int argc, char* argv[]) {
 		}
 	}
 
-	serial = serial_open(device);
-	if (!serial) {
-		fprintf(stderr, "Failed to open serial port: ");
-		perror(device);
+	if (port_open(&port_opts, &port) != PORT_ERR_OK) {
+		fprintf(stderr, "Failed to open port: %s\n", port_opts.device);
 		goto close;
 	}
 
-	if (serial_setup(
-		serial,
-		baudRate,
-		serial_get_bits(serial_mode),
-		serial_get_parity(serial_mode),
-		serial_get_stopbit(serial_mode)
-	) != SERIAL_ERR_OK) {
-		perror(device);
+	fprintf(diag, "Interface %s: %s\n", port->name, port->get_cfg_str(port));
+	if (init_flag && init_bl_entry(port, gpio_seq) == 0)
+		goto close;
+	stm = stm32_init(port, init_flag);
+	if (!stm)
 		goto close;
-	}
-
-	fprintf(diag, "Serial Config: %s\n", serial_get_setup_str(serial));
-	if (init_flag && init_bl_entry(serial, gpio_seq) == 0) goto close;
-	if (!(stm = stm32_init(serial, init_flag))) goto close;
 
 	fprintf(diag, "Version      : 0x%02x\n", stm->bl_version);
-	fprintf(diag, "Option 1     : 0x%02x\n", stm->option1);
-	fprintf(diag, "Option 2     : 0x%02x\n", stm->option2);
+	if (port->flags & PORT_GVR_ETX) {
+		fprintf(diag, "Option 1     : 0x%02x\n", stm->option1);
+		fprintf(diag, "Option 2     : 0x%02x\n", stm->option2);
+	}
 	fprintf(diag, "Device ID    : 0x%04x (%s)\n", stm->pid, stm->dev->name);
 	fprintf(diag, "- RAM        : %dKiB  (%db reserved by bootloader)\n", (stm->dev->ram_end - 0x20000000) / 1024, stm->dev->ram_start - 0x20000000);
 	fprintf(diag, "- Flash      : %dKiB (sector size: %dx%d)\n", (stm->dev->fl_end - stm->dev->fl_start ) / 1024, stm->dev->fl_pps, stm->dev->fl_ps);
@@ -169,38 +202,84 @@ int main(int argc, char* argv[]) {
 	uint32_t	addr, start, end;
 	unsigned int	len;
 	int		failed = 0;
+	int		first_page, num_pages;
 
-	if (rd) {
-		fprintf(diag, "\n");
+	/*
+	 * Cleanup addresses:
+	 *
+	 * Starting from options
+	 *	start_addr, readwrite_len, spage, npages
+	 * and using device memory size, compute
+	 *	start, end, first_page, num_pages
+	 */
+	if (start_addr || readwrite_len) {
+		start = start_addr;
 
-		if ((perr = parser->open(p_st, filename, 1)) != PARSER_ERR_OK) {
-			fprintf(stderr, "%s ERROR: %s\n", parser->name, parser_errstr(perr));
-			if (perr == PARSER_ERR_SYSTEM) perror(filename);
+		if (is_addr_in_flash(start))
+			end = stm->dev->fl_end;
+		else {
+			no_erase = 1;
+			if (is_addr_in_ram(start))
+				end = stm->dev->ram_end;
+			else
+				end = start + sizeof(uint32_t);
+		}
+
+		if (readwrite_len && (end > start + readwrite_len))
+			end = start + readwrite_len;
+
+		first_page = flash_addr_to_page_floor(start);
+		if (!first_page && end == stm->dev->fl_end)
+			num_pages = 0xff; /* mass erase */
+		else
+			num_pages = flash_addr_to_page_ceil(end) - first_page;
+	} else if (!spage && !npages) {
+		start = stm->dev->fl_start;
+		end = stm->dev->fl_end;
+		first_page = 0;
+		num_pages = 0xff; /* mass erase */
+	} else {
+		first_page = spage;
+		start = flash_page_to_addr(first_page);
+		if (start > stm->dev->fl_end) {
+			fprintf(stderr, "Address range exceeds flash size.\n");
 			goto close;
 		}
 
-		if (start_addr || readwrite_len) {
-			start = start_addr;
-			if (readwrite_len)
-				end = start_addr + readwrite_len;
-			else
+		if (npages) {
+			num_pages = npages;
+			end = flash_page_to_addr(first_page + num_pages);
+			if (end > stm->dev->fl_end)
 				end = stm->dev->fl_end;
 		} else {
-			start = stm->dev->fl_start + (spage * stm->dev->fl_ps);
 			end = stm->dev->fl_end;
+			num_pages = flash_addr_to_page_ceil(end) - first_page;
 		}
-		addr = start;
 
-		if (start < stm->dev->fl_start || end > stm->dev->fl_end) {
-			fprintf(stderr, "Specified start & length are invalid\n");
+		if (!first_page && end == stm->dev->fl_end)
+			num_pages = 0xff; /* mass erase */
+	}
+
+	if (rd) {
+		unsigned int max_len = port_opts.rx_frame_max;
+
+		fprintf(diag, "Memory read\n");
+
+		perr = parser->open(p_st, filename, 1);
+		if (perr != PARSER_ERR_OK) {
+			fprintf(stderr, "%s ERROR: %s\n", parser->name, parser_errstr(perr));
+			if (perr == PARSER_ERR_SYSTEM)
+				perror(filename);
 			goto close;
 		}
 
 		fflush(diag);
+		addr = start;
 		while(addr < end) {
 			uint32_t left	= end - addr;
-			len		= sizeof(buffer) > left ? left : sizeof(buffer);
-			if (!stm32_read_memory(stm, addr, buffer, len)) {
+			len		= max_len > left ? left : max_len;
+			s_err = stm32_read_memory(stm, addr, buffer, len);
+			if (s_err != STM32_ERR_OK) {
 				fprintf(stderr, "Failed to read memory at address 0x%08x, target write-protected?\n", addr);
 				goto close;
 			}
@@ -236,29 +315,21 @@ int main(int argc, char* argv[]) {
 	} else if (eraseOnly) {
 		ret = 0;
 		fprintf(stdout, "Erasing flash\n");
-		if (start_addr || readwrite_len) {
-			if ((start_addr % stm->dev->fl_ps) != 0
-			    || (readwrite_len % stm->dev->fl_ps) != 0) {
-				fprintf(stderr, "Specified start & length are invalid (must be page aligned)\n");
-				ret = 1;
-				goto close;
-			}
-			spage = (start_addr - stm->dev->fl_start) / stm->dev->fl_ps;
-			if (readwrite_len)
-				npages = readwrite_len / stm->dev->fl_ps;
-			else
-				npages = (stm->dev->fl_end - stm->dev->fl_start) / stm->dev->fl_ps;
+
+		if (num_pages != 0xff &&
+		    (start != flash_page_to_addr(first_page)
+		     || end != flash_page_to_addr(first_page + num_pages))) {
+			fprintf(stderr, "Specified start & length are invalid (must be page aligned)\n");
+			ret = 1;
+			goto close;
 		}
 
-		if (!spage && !npages)
-			npages = 0xff; /* mass erase */
-
-		if (!stm32_erase_memory(stm, spage, npages)) {
+		s_err = stm32_erase_memory(stm, first_page, num_pages);
+		if (s_err != STM32_ERR_OK) {
 			fprintf(stderr, "Failed to erase memory\n");
 			ret = 1;
 			goto close;
 		}
-		
 	} else if (wu) {
 		fprintf(diag, "Write-unprotecting flash\n");
 		/* the device automatically performs a reset after the sending the ACK */
@@ -267,49 +338,25 @@ int main(int argc, char* argv[]) {
 		fprintf(diag,	"Done.\n");
 
 	} else if (wr) {
-		fprintf(diag, "\n");
+		fprintf(diag, "Write to memory\n");
 
 		off_t 	offset = 0;
 		ssize_t r;
 		unsigned int size;
+		unsigned int max_wlen, max_rlen;
+
+		max_wlen = port_opts.tx_frame_max - 2;	/* skip len and crc */
+		max_wlen &= ~3;	/* 32 bit aligned */
+
+		max_rlen = port_opts.rx_frame_max;
+		max_rlen = max_rlen < max_wlen ? max_rlen : max_wlen;
 
 		/* Assume data from stdin is whole device */
-		if (filename[0] == '-')
-			size = stm->dev->fl_end - stm->dev->fl_start;
+		if (filename[0] == '-' && filename[1] == '\0')
+			size = end - start;
 		else
 			size = parser->size(p_st);
 
-		if (start_addr || readwrite_len) {
-			start = start_addr;
-			spage = (start_addr - stm->dev->fl_start) / stm->dev->fl_ps;
-			if (readwrite_len) {
-				end = start_addr + readwrite_len;
-				npages = (end - stm->dev->fl_start + stm->dev->fl_ps - 1) / stm->dev->fl_ps - spage;
-			} else {
-				end = stm->dev->fl_end;
-				if (spage)
-					npages = (end - stm->dev->fl_start) / stm->dev->fl_ps - spage;
-				else
-					npages = 0xff; /* mass erase */
-			}
-		} else if (!spage && !npages) {
-			start = stm->dev->fl_start;
-			end = stm->dev->fl_end;
-			npages = 0xff; /* mass erase */
-		} else {
-			start = stm->dev->fl_start + (spage * stm->dev->fl_ps);
-			if (npages)
-				end = start + npages * stm->dev->fl_ps;
-			else
-				end = stm->dev->fl_end;
-		}
-		addr = start;
-
-		if (start < stm->dev->fl_start || end > stm->dev->fl_end) {
-			fprintf(stderr, "Specified start & length are invalid\n");
-			goto close;
-		}
-
 		// TODO: It is possible to write to non-page boundaries, by reading out flash
 		//       from partial pages and combining with the input data
 		// if ((start % stm->dev->fl_ps) != 0 || (end % stm->dev->fl_ps) != 0) {
@@ -319,15 +366,20 @@ int main(int argc, char* argv[]) {
 
 		// TODO: If writes are not page aligned, we should probably read out existing flash
 		//       contents first, so it can be preserved and combined with new data
-		if (!stm32_erase_memory(stm, spage, npages)) {
-			fprintf(stderr, "Failed to erase memory\n");
-			goto close;
+		if (!no_erase && num_pages) {
+			fprintf(diag, "Erasing memory\n");
+			s_err = stm32_erase_memory(stm, first_page, num_pages);
+			if (s_err != STM32_ERR_OK) {
+				fprintf(stderr, "Failed to erase memory\n");
+				goto close;
+			}
 		}
 
 		fflush(diag);
+		addr = start;
 		while(addr < end && offset < size) {
 			uint32_t left	= end - addr;
-			len		= sizeof(buffer) > left ? left : sizeof(buffer);
+			len		= max_wlen > left ? left : max_wlen;
 			len		= len > size - offset ? size - offset : len;
 
 			if (parser->read(p_st, buffer, &len) != PARSER_ERR_OK)
@@ -343,16 +395,26 @@ int main(int argc, char* argv[]) {
 			}
 	
 			again:
-			if (!stm32_write_memory(stm, addr, buffer, len)) {
+			s_err = stm32_write_memory(stm, addr, buffer, len);
+			if (s_err != STM32_ERR_OK) {
 				fprintf(stderr, "Failed to write memory at address 0x%08x\n", addr);
 				goto close;
 			}
 
 			if (verify) {
 				uint8_t compare[len];
-				if (!stm32_read_memory(stm, addr, compare, len)) {
-					fprintf(stderr, "Failed to read memory at address 0x%08x\n", addr);
-					goto close;
+				unsigned int offset, rlen;
+
+				offset = 0;
+				while (offset < len) {
+					rlen = len - offset;
+					rlen = rlen < max_rlen ? rlen : max_rlen;
+					s_err = stm32_read_memory(stm, addr + offset, compare + offset, rlen);
+					if (s_err != STM32_ERR_OK) {
+						fprintf(stderr, "Failed to read memory at address 0x%08x\n", addr + offset);
+						goto close;
+					}
+					offset += rlen;
 				}
 
 				for(r = 0; r < len; ++r)
@@ -388,6 +450,20 @@ int main(int argc, char* argv[]) {
 		fprintf(diag,	"Done.\n");
 		ret = 0;
 		goto close;
+	} else if (crc) {
+		uint32_t crc_val = 0;
+
+		fprintf(diag, "CRC computation\n");
+
+		s_err = stm32_crc_wrapper(stm, start, end - start, &crc_val);
+		if (s_err != STM32_ERR_OK) {
+			fprintf(stderr, "Failed to read CRC\n");
+			goto close;
+		}
+		fprintf(diag, "CRC(0x%08x-0x%08x) = 0x%08x\n", start, end,
+			crc_val);
+		ret = 0;
+		goto close;
 	} else
 		ret = 0;
 
@@ -398,7 +474,7 @@ close:
 
 		fprintf(diag, "\nStarting execution at address 0x%08x... ", execute);
 		fflush(diag);
-		if (stm32_go(stm, execute)) {
+		if (stm32_go(stm, execute) == STM32_ERR_OK) {
 			reset_flag = 0;
 			fprintf(diag, "done.\n");
 		} else
@@ -408,29 +484,38 @@ close:
 	if (stm && reset_flag) {
 		fprintf(diag, "\nResetting device... ");
 		fflush(diag);
-		if (init_bl_exit(stm, serial, gpio_seq))
+		if (init_bl_exit(stm, port, gpio_seq))
 			fprintf(diag, "done.\n");
 		else	fprintf(diag, "failed.\n");
 	}
 
 	if (p_st  ) parser->close(p_st);
 	if (stm   ) stm32_close  (stm);
-	if (serial) serial_close (serial);
+	if (port)
+		port->close(port);
 
 	fprintf(diag, "\n");
 	return ret;
 }
 
-int parse_options(int argc, char *argv[]) {
+int parse_options(int argc, char *argv[])
+{
 	int c;
-	while((c = getopt(argc, argv, "b:m:r:w:e:vn:g:jkfchuos:S:i:")) != -1) {
+	char *pLen;
+
+	while ((c = getopt(argc, argv, "a:b:m:r:w:e:vn:g:jkfcChuos:S:F:i:R")) != -1) {
 		switch(c) {
+			case 'a':
+				port_opts.bus_addr = strtoul(optarg, NULL, 0);
+				break;
+
 			case 'b':
-				baudRate = serial_get_baud(strtoul(optarg, NULL, 0));
-				if (baudRate == SERIAL_BAUD_INVALID) {
+				port_opts.baudRate = serial_get_baud(strtoul(optarg, NULL, 0));
+				if (port_opts.baudRate == SERIAL_BAUD_INVALID) {
+					serial_baud_t baudrate;
 					fprintf(stderr,	"Invalid baud rate, valid options are:\n");
-					for(baudRate = SERIAL_BAUD_1200; baudRate != SERIAL_BAUD_INVALID; ++baudRate)
-						fprintf(stderr, " %d\n", serial_get_baud_int(baudRate));
+					for (baudrate = SERIAL_BAUD_1200; baudrate != SERIAL_BAUD_INVALID; ++baudrate)
+						fprintf(stderr, " %d\n", serial_get_baud_int(baudrate));
 					return 1;
 				}
 				break;
@@ -443,7 +528,7 @@ int parse_options(int argc, char *argv[]) {
 					fprintf(stderr, "Invalid serial mode\n");
 					return 1;
 				}
-				serial_mode = optarg;
+				port_opts.serial_mode = optarg;
 				break;
 
 			case 'r':
@@ -469,6 +554,8 @@ int parse_options(int argc, char *argv[]) {
 					fprintf(stderr, "ERROR: You need to specify a page count between 0 and 255");
 					return 1;
 				}
+				if (!npages)
+					no_erase = 1;
 				break;
 			case 'u':
 				wu = 1;
@@ -530,7 +617,6 @@ int parse_options(int argc, char *argv[]) {
 					fprintf(stderr, "ERROR: Invalid options, can't specify start page / num pages and start address/length\n");
 					return 1;
 				} else {
-					char *pLen;
 					start_addr = strtoul(optarg, &pLen, 0);
 					if (*pLen == ':') {
 						pLen++;
@@ -542,6 +628,36 @@ int parse_options(int argc, char *argv[]) {
 					}
 				}
 				break;
+			case 'F':
+				port_opts.rx_frame_max = strtoul(optarg, &pLen, 0);
+				if (*pLen == ':') {
+					pLen++;
+					port_opts.tx_frame_max = strtoul(pLen, NULL, 0);
+				}
+				if (port_opts.rx_frame_max < 0
+				    || port_opts.tx_frame_max < 0) {
+					fprintf(stderr, "ERROR: Invalid negative value for option -F\n");
+					return 1;
+				}
+				if (port_opts.rx_frame_max == 0)
+					port_opts.rx_frame_max = STM32_MAX_RX_FRAME;
+				if (port_opts.tx_frame_max == 0)
+					port_opts.tx_frame_max = STM32_MAX_TX_FRAME;
+				if (port_opts.rx_frame_max < 20
+				    || port_opts.tx_frame_max < 5) {
+					fprintf(stderr, "ERROR: current code cannot work with small frames.\n");
+					fprintf(stderr, "min(RX) = 20, min(TX) = 5\n");
+					return 1;
+				}
+				if (port_opts.rx_frame_max > STM32_MAX_RX_FRAME) {
+					fprintf(stderr, "WARNING: Ignore RX length in option -F\n");
+					port_opts.rx_frame_max = STM32_MAX_RX_FRAME;
+				}
+				if (port_opts.tx_frame_max > STM32_MAX_TX_FRAME) {
+					fprintf(stderr, "WARNING: Ignore TX length in option -F\n");
+					port_opts.tx_frame_max = STM32_MAX_TX_FRAME;
+				}
+				break;
 			case 'f':
 				force_binary = 1;
 				break;
@@ -557,19 +673,27 @@ int parse_options(int argc, char *argv[]) {
 			case 'i':
 				gpio_seq = optarg;
 				break;
+
+			case 'R':
+				reset_flag = 1;
+				break;
+
+			case 'C':
+				crc = 1;
+				break;
 		}
 	}
 
 	for (c = optind; c < argc; ++c) {
-		if (device) {
+		if (port_opts.device) {
 			fprintf(stderr, "ERROR: Invalid parameter specified\n");
 			show_help(argv[0]);
 			return 1;
 		}
-		device = argv[c];
+		port_opts.device = argv[c];
 	}
 
-	if (device == NULL) {
+	if (port_opts.device == NULL) {
 		fprintf(stderr, "ERROR: Device not specified\n");
 		show_help(argv[0]);
 		return 1;
@@ -586,11 +710,13 @@ int parse_options(int argc, char *argv[]) {
 
 void show_help(char *name) {
 	fprintf(stderr,
-		"Usage: %s [-bvngfhc] [-[rw] filename] /dev/ttyS0\n"
+		"Usage: %s [-bvngfhc] [-[rw] filename] [tty_device | i2c_device]\n"
+		"	-a bus_address	Bus address (e.g. for I2C port)\n"
 		"	-b rate		Baud rate (default 57600)\n"
 		"	-m mode		Serial port mode (default 8e1)\n"
 		"	-r filename	Read flash to file (or - stdout)\n"
 		"	-w filename	Write flash from file (or - stdout)\n"
+		"	-C		Compute CRC of flash content\n"
 		"	-u		Disable the flash write-protection\n"
 		"	-j		Enable the flash read-protection\n"
 		"	-k		Disable the flash read-protection\n"
@@ -601,6 +727,7 @@ void show_help(char *name) {
 		"	-g address	Start execution at specified address (0 = flash start)\n"
 		"	-S address[:length]	Specify start address and optionally length for\n"
 		"	                   	read/write/erase operations\n"
+		"	-F RX_length[:TX_length]  Specify the max length of RX and TX frame\n"
 		"	-s start_page	Flash at specified page (0 = flash start)\n"
 		"	-f		Force binary parser\n"
 		"	-h		Show this help\n"
@@ -610,10 +737,13 @@ void show_help(char *name) {
 		"	-i GPIO_string	GPIO sequence to enter/exit bootloader mode\n"
 		"			GPIO_string=[entry_seq][:[exit_seq]]\n"
 		"			sequence=[-]n[,sequence]\n"
+		"	-R		Reset device at exit.\n"
 		"\n"
 		"Examples:\n"
 		"	Get device information:\n"
 		"		%s /dev/ttyS0\n"
+		"	  or:\n"
+		"		%s /dev/i2c-0\n"
 		"\n"
 		"	Write with verify and then start execution:\n"
 		"		%s -w filename -v -g 0x0 /dev/ttyS0\n"
@@ -637,6 +767,7 @@ void show_help(char *name) {
 		name,
 		name,
 		name,
+		name,
 		name
 	);
 }
